@@ -58,7 +58,7 @@
 
 #endif /* HAVE_CLUTTER_GTK_{X11,WIN32} */
 
-G_DEFINE_TYPE (GtkClutterEmbed, gtk_clutter_embed, GTK_TYPE_WIDGET);
+G_DEFINE_TYPE (GtkClutterEmbed, gtk_clutter_embed, GTK_TYPE_CONTAINER);
 
 #define GTK_CLUTTER_EMBED_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GTK_CLUTTER_TYPE_EMBED, GtkClutterEmbedPrivate))
 
@@ -72,17 +72,19 @@ struct _GtkClutterEmbedPrivate
 static void
 gtk_clutter_embed_send_configure (GtkClutterEmbed *embed)
 {
-  GtkWidget *widget;
   GdkEvent *event = gdk_event_new (GDK_CONFIGURE);
+  GtkAllocation allocation;
+  GtkWidget *widget;
 
   widget = GTK_WIDGET (embed);
+  gtk_widget_get_allocation (widget, &allocation);
 
-  event->configure.window = g_object_ref (widget->window);
+  event->configure.window = g_object_ref (gtk_widget_get_window (widget));
   event->configure.send_event = TRUE;
-  event->configure.x = widget->allocation.x;
-  event->configure.y = widget->allocation.y;
-  event->configure.width = widget->allocation.width;
-  event->configure.height = widget->allocation.height;
+  event->configure.x = allocation.x;
+  event->configure.y = allocation.y;
+  event->configure.width = allocation.width;
+  event->configure.height = allocation.height;
   
   gtk_widget_event (widget, event);
   gdk_event_free (event);
@@ -131,7 +133,7 @@ gtk_clutter_embed_show (GtkWidget *widget)
 {
   GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
 
-  if (GTK_WIDGET_REALIZED (widget))
+  if (gtk_widget_get_realized (widget))
     clutter_actor_show (priv->stage);
 
   GTK_WIDGET_CLASS (gtk_clutter_embed_parent_class)->show (widget);
@@ -155,7 +157,10 @@ gtk_clutter_embed_realize (GtkWidget *widget)
 {
   GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv; 
   GdkWindowAttr attributes;
+  GtkAllocation allocation;
   int attributes_mask;
+  GdkWindow *window;
+  GtkStyle *style;
 
 #ifdef HAVE_CLUTTER_GTK_X11
   {
@@ -164,7 +169,7 @@ gtk_clutter_embed_realize (GtkWidget *widget)
     GdkColormap *colormap;
 
     /* We need to use the colormap from the Clutter visual */
-    xvinfo = clutter_x11_get_stage_visual (CLUTTER_STAGE (priv->stage));
+    xvinfo = clutter_x11_get_visual_info ();
     if (xvinfo == None)
       {
         g_critical ("Unable to retrieve the XVisualInfo from Clutter");
@@ -176,15 +181,18 @@ gtk_clutter_embed_realize (GtkWidget *widget)
     colormap = gdk_colormap_new (visual, FALSE);
     gtk_widget_set_colormap (widget, colormap);
   }
-#endif
+#endif /* HAVE_CLUTTER_GTK_X11 */
 
-  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
+  gtk_widget_set_realized (widget, TRUE);
 
   attributes.window_type = GDK_WINDOW_CHILD;
-  attributes.x = widget->allocation.x;
-  attributes.y = widget->allocation.y;
-  attributes.width = widget->allocation.width;
-  attributes.height = widget->allocation.height;
+
+  gtk_widget_get_allocation (widget, &allocation);
+  attributes.x = allocation.x;
+  attributes.y = allocation.y;
+  attributes.width = allocation.width;
+  attributes.height = allocation.height;
+
   attributes.wclass = GDK_INPUT_OUTPUT;
   attributes.visual = gtk_widget_get_visual (widget);
   attributes.colormap = gtk_widget_get_colormap (widget);
@@ -198,31 +206,34 @@ gtk_clutter_embed_realize (GtkWidget *widget)
                         | GDK_BUTTON_RELEASE_MASK
                         | GDK_KEY_PRESS_MASK
                         | GDK_KEY_RELEASE_MASK
+                        | GDK_ENTER_NOTIFY_MASK
+                        | GDK_LEAVE_NOTIFY_MASK
                         | GDK_POINTER_MOTION_MASK;
 
   attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
 
-  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
-                                   &attributes,
-                                   attributes_mask);
-  gdk_window_set_user_data (widget->window, widget);
+  window = gdk_window_new (gtk_widget_get_parent_window (widget),
+                           &attributes,
+                           attributes_mask);
+  gdk_window_set_user_data (window, widget);
+  gdk_window_set_back_pixmap (window, NULL, FALSE);
+  gtk_widget_set_window (widget, window);
 
-  widget->style = gtk_style_attach (widget->style, widget->window);
-  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);
+  gtk_widget_style_attach (widget);
+  style = gtk_widget_get_style (widget);
+  gtk_style_set_background (style, window, GTK_STATE_NORMAL);
   
-  gdk_window_set_back_pixmap (widget->window, NULL, FALSE);
-
 #if defined(HAVE_CLUTTER_GTK_X11)
   clutter_x11_set_stage_foreign (CLUTTER_STAGE (priv->stage), 
-                                 GDK_WINDOW_XID (widget->window));
+                                 GDK_WINDOW_XID (window));
 #elif defined(HAVE_CLUTTER_GTK_WIN32)
   clutter_win32_set_stage_foreign (CLUTTER_STAGE (priv->stage), 
-				   GDK_WINDOW_HWND (widget->window));
+				   GDK_WINDOW_HWND (window));
 #endif /* HAVE_CLUTTER_GTK_{X11,WIN32} */
 
   clutter_actor_realize (priv->stage);
 
-  if (GTK_WIDGET_VISIBLE (widget))
+  if (gtk_widget_get_visible (widget))
     clutter_actor_show (priv->stage);
 
   gtk_clutter_embed_send_configure (GTK_CLUTTER_EMBED (widget));
@@ -244,13 +255,15 @@ gtk_clutter_embed_size_allocate (GtkWidget     *widget,
 {
   GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
 
-  widget->allocation = *allocation;
+  gtk_widget_set_allocation (widget, allocation);
 
-  if (GTK_WIDGET_REALIZED (widget))
+  if (gtk_widget_get_realized (widget))
     {
-      gdk_window_move_resize (widget->window,
-                              allocation->x, allocation->y,
-                              allocation->width, allocation->height);
+      gdk_window_move_resize (gtk_widget_get_window (widget),
+                              allocation->x,
+                              allocation->y,
+                              allocation->width,
+                              allocation->height);
 
       gtk_clutter_embed_send_configure (GTK_CLUTTER_EMBED (widget));
     }
@@ -282,29 +295,38 @@ gtk_clutter_embed_motion_notify_event (GtkWidget      *widget,
                                        GdkEventMotion *event)
 {
   GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
+  ClutterDeviceManager *manager;
+  ClutterInputDevice *device;
   ClutterEvent cevent = { 0, };
 
-  cevent.type = CLUTTER_MOTION;
-  cevent.any.stage = CLUTTER_STAGE (priv->stage);
+  manager = clutter_device_manager_get_default ();
+  device = clutter_device_manager_get_core_device (manager, CLUTTER_POINTER_DEVICE);
+
+  cevent.motion.type = CLUTTER_MOTION;
+  cevent.motion.stage = CLUTTER_STAGE (priv->stage);
   cevent.motion.x = event->x;
   cevent.motion.y = event->y;
   cevent.motion.time = event->time;
   cevent.motion.modifier_state = event->state;
+  cevent.motion.device = device;
+
+  clutter_input_device_update_from_event (device, &cevent, FALSE);
 
   clutter_do_event (&cevent);
 
-  /* doh - motion events can push ENTER/LEAVE events onto Clutters
-   * internal event queue which we do really ever touch (essentially
-   * proxying from gtks queue). The below pumps them back out and
-   * processes.
-   * *could* be side effects with below though doubful as no other
-   * events reach the queue (we shut down event collection). Maybe
-   * a peek_mask type call could be even safer. 
-  */
-  while (clutter_events_pending())
+  /* XXX - motion events might synthesize ENTER/LEAVE event pairs
+   * and push them on the Clutter event queue for our stage; we need
+   * to pop them out of the queue.
+   *
+   * FIXME - instead of using the global "are there events pending"
+   * function we should have a clutter_stage_events_pending()
+   * accessor so that we can distinguish between stages.
+   */
+  while (clutter_events_pending ())
     {
       ClutterEvent *ev = clutter_event_get ();
-      if (ev)
+
+      if (ev != NULL)
         {
           clutter_do_event (ev);
           clutter_event_free (ev);
@@ -319,18 +341,23 @@ gtk_clutter_embed_button_event (GtkWidget      *widget,
                                 GdkEventButton *event)
 {
   GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
+  ClutterDeviceManager *manager;
+  ClutterInputDevice *device;
   ClutterEvent cevent = { 0, };
 
   if (event->type == GDK_BUTTON_PRESS ||
       event->type == GDK_2BUTTON_PRESS ||
       event->type == GDK_3BUTTON_PRESS)
-    cevent.type = cevent.button.type = CLUTTER_BUTTON_PRESS;
+    cevent.button.type = CLUTTER_BUTTON_PRESS;
   else if (event->type == GDK_BUTTON_RELEASE)
-    cevent.type = cevent.button.type = CLUTTER_BUTTON_RELEASE;
+    cevent.button.type = CLUTTER_BUTTON_RELEASE;
   else
     return FALSE;
 
-  cevent.any.stage = CLUTTER_STAGE (priv->stage);
+  manager = clutter_device_manager_get_default ();
+  device = clutter_device_manager_get_core_device (manager, CLUTTER_POINTER_DEVICE);
+
+  cevent.button.stage = CLUTTER_STAGE (priv->stage);
   cevent.button.x = event->x;
   cevent.button.y = event->y;
   cevent.button.time = event->time;
@@ -340,6 +367,9 @@ gtk_clutter_embed_button_event (GtkWidget      *widget,
                                                                          : 3));
   cevent.button.modifier_state = event->state;
   cevent.button.button = event->button;
+  cevent.button.device = device;
+
+  clutter_input_device_update_from_event (device, &cevent, FALSE);
 
   clutter_do_event (&cevent);
 
@@ -351,21 +381,27 @@ gtk_clutter_embed_key_event (GtkWidget   *widget,
                              GdkEventKey *event)
 {
   GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
+  ClutterDeviceManager *manager;
+  ClutterInputDevice *device;
   ClutterEvent cevent = { 0, };
 
   if (event->type == GDK_KEY_PRESS)
-    cevent.type = cevent.key.type = CLUTTER_KEY_PRESS;
+    cevent.key.type = CLUTTER_KEY_PRESS;
   else if (event->type == GDK_KEY_RELEASE)
-    cevent.type = cevent.key.type = CLUTTER_KEY_RELEASE;
+    cevent.key.type = CLUTTER_KEY_RELEASE;
   else
     return FALSE;
 
-  cevent.any.stage = CLUTTER_STAGE (priv->stage);
+  manager = clutter_device_manager_get_default ();
+  device = clutter_device_manager_get_core_device (manager, CLUTTER_KEYBOARD_DEVICE);
+
+  cevent.key.stage = CLUTTER_STAGE (priv->stage);
   cevent.key.time = event->time;
   cevent.key.modifier_state = event->state;
   cevent.key.keyval = event->keyval;
   cevent.key.hardware_keycode = event->hardware_keycode;
   cevent.key.unicode_value = gdk_keyval_to_unicode (event->keyval);
+  cevent.key.device = device;
 
   clutter_do_event (&cevent);
 
@@ -466,6 +502,62 @@ gtk_clutter_embed_scroll_event (GtkWidget      *widget,
   return FALSE;
 }
 
+static gboolean
+gtk_clutter_embed_enter_notify_event (GtkWidget        *widget,
+                                      GdkEventCrossing *event)
+{
+  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
+  ClutterDeviceManager *manager;
+  ClutterInputDevice *device;
+  ClutterEvent cevent = { 0, };
+
+  manager = clutter_device_manager_get_default ();
+  device = clutter_device_manager_get_core_device (manager, CLUTTER_POINTER_DEVICE);
+
+  cevent.crossing.type = CLUTTER_ENTER;
+  cevent.crossing.stage = CLUTTER_STAGE (priv->stage);
+  cevent.crossing.x = (gint) event->x;
+  cevent.crossing.y = (gint) event->y;
+  cevent.crossing.time = event->time;
+  cevent.crossing.source = priv->stage;
+  cevent.crossing.related = NULL;
+  cevent.crossing.device = device;
+
+  clutter_input_device_update_from_event (device, &cevent, TRUE);
+
+  clutter_do_event (&cevent);
+
+  return FALSE;
+}
+
+static gboolean
+gtk_clutter_embed_leave_notify_event (GtkWidget        *widget,
+                                      GdkEventCrossing *event)
+{
+  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
+  ClutterDeviceManager *manager;
+  ClutterInputDevice *device;
+  ClutterEvent cevent = { 0, };
+
+  manager = clutter_device_manager_get_default ();
+  device = clutter_device_manager_get_core_device (manager, CLUTTER_POINTER_DEVICE);
+
+  cevent.crossing.type = CLUTTER_LEAVE;
+  cevent.crossing.stage = CLUTTER_STAGE (priv->stage);
+  cevent.crossing.x = (gint) event->x;
+  cevent.crossing.y = (gint) event->y;
+  cevent.crossing.time = event->time;
+  cevent.crossing.source = priv->stage;
+  cevent.crossing.related = NULL;
+  cevent.crossing.device = device;
+
+  clutter_input_device_update_from_event (device, &cevent, TRUE);
+
+  clutter_do_event (&cevent);
+
+  return FALSE;
+}
+
 static void
 gtk_clutter_embed_style_set (GtkWidget *widget,
                              GtkStyle  *old_style)
@@ -513,11 +605,41 @@ gtk_clutter_embed_style_set (GtkWidget *widget,
   g_free (font_name);
 }
 
+static void
+gtk_clutter_embed_add (GtkContainer	 *container,
+		       GtkWidget	 *widget)
+{
+  g_warning ("GtkClutterEmbed children not yet supported");
+}
+
+static void
+gtk_clutter_embed_remove (GtkContainer	 *container,
+			  GtkWidget	 *widget)
+{
+  g_warning ("GtkClutterEmbed children not yet supported");
+}
+
+static void
+gtk_clutter_embed_forall (GtkContainer	 *container,
+			  gboolean	  include_internals,
+			  GtkCallback	  callback,
+			  gpointer	  callback_data)
+{
+}
+
+static GType
+gtk_clutter_embed_child_type (GtkContainer *container)
+{
+  /* No children supported yet */
+  return G_TYPE_NONE;
+}
+
 static void
 gtk_clutter_embed_class_init (GtkClutterEmbedClass *klass)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
+  GtkContainerClass *container_class = GTK_CONTAINER_CLASS (klass);
 
   g_type_class_add_private (klass, sizeof (GtkClutterEmbedPrivate));
 
@@ -542,17 +664,25 @@ gtk_clutter_embed_class_init (GtkClutterEmbedClass *klass)
   widget_class->focus_in_event = gtk_clutter_embed_focus_in;
   widget_class->focus_out_event = gtk_clutter_embed_focus_out;
   widget_class->scroll_event = gtk_clutter_embed_scroll_event;
+  widget_class->enter_notify_event = gtk_clutter_embed_enter_notify_event;
+  widget_class->leave_notify_event = gtk_clutter_embed_leave_notify_event;
+
+  container_class->add = gtk_clutter_embed_add;
+  container_class->remove = gtk_clutter_embed_remove;
+  container_class->forall = gtk_clutter_embed_forall;
+  container_class->child_type = gtk_clutter_embed_child_type;
 }
 
 static void
 gtk_clutter_embed_init (GtkClutterEmbed *embed)
 {
+  GtkWidget *widget = GTK_WIDGET (embed);
   GtkClutterEmbedPrivate *priv;
 
   embed->priv = priv = GTK_CLUTTER_EMBED_GET_PRIVATE (embed);
 
-  GTK_WIDGET_SET_FLAGS (embed, GTK_CAN_FOCUS);
-  GTK_WIDGET_UNSET_FLAGS (embed, GTK_NO_WINDOW);
+  gtk_widget_set_can_focus (widget, TRUE);
+  gtk_widget_set_has_window (widget, TRUE);
 
   /* disable double-buffering: it's automatically provided
    * by OpenGL
@@ -593,10 +723,11 @@ gtk_clutter_embed_new (void)
  * @embed: a #GtkClutterEmbed
  *
  * Retrieves the #ClutterStage from @embed. The returned stage can be
- * used to add actors to the Clutter scene.
+ * used to add actors to the Clutter scene. Multiple calls to this function
+ * on the same #GtkClutterEmbed widget will return the same stage.
  *
- * Return value: the Clutter stage. You should never destroy or unref
- *   the returned actor.
+ * Return value: (transfer none): a #ClutterStage. You should never destroy
+ *   or unref the returned actor.
  *
  * Since: 0.6
  */
